from unittest import TestCase

from regret import EmittedDeprecation, Deprecator
from regret.testing import Recorder


def calculate():
    """
    Perform a super important calculation.
    """
    return 12


def add(x, y):
    return 12


class TestRegret(TestCase):
    def setUp(self):
        self.recorder = Recorder()
        self.regret = Deprecator(emit=self.recorder.emit)

    def test_function(self):
        self.assertEqual(
            (self.regret.callable()(calculate)(), self.recorder), (
                12,
                Recorder(saw=[EmittedDeprecation(object=calculate)]),
            ),
        )

    def test_function_with_args(self):
        self.assertEqual(
            (self.regret.callable()(add)(9, y=3), self.recorder), (
                12,
                Recorder(saw=[EmittedDeprecation(object=add)]),
            ),
        )

    def test_function_is_wrapped(self):
        deprecated = self.regret.callable()(calculate)
        self.assertEqual(
            (calculate.__name__, calculate.__doc__),
            (deprecated.__name__, deprecated.__doc__),
        )

    def test_method(self):
        class Calculator(object):
            def _calculate(self):
                return 12

            calculate = self.regret.callable()(_calculate)

        unbound = getattr(
            Calculator._calculate, "im_func", Calculator._calculate,
        )

        self.assertEqual(
            (Calculator().calculate(), self.recorder), (
                12,
                Recorder(saw=[EmittedDeprecation(object=unbound)]),
            ),
        )

    def test_method_is_wrapped(self):
        class Calculator(object):
            def _calculate(self):
                """
                Perform a super important calculation.
                """
                return 12

            calculate = self.regret.callable()(_calculate)

        self.assertEqual(
            (
                Calculator.calculate.__name__,
                Calculator.calculate.__doc__,
                Calculator().calculate.__name__,
                Calculator().calculate.__doc__,
            ),
            (
                Calculator._calculate.__name__,
                Calculator._calculate.__doc__,
                Calculator._calculate.__name__,
                Calculator._calculate.__doc__,
            ),
        )

    def test_dunder_call(self):
        class Calculator(object):
            def _calculate(self):
                return 12

            __call__ = self.regret.callable()(_calculate)

        unbound = getattr(
            Calculator._calculate, "im_func", Calculator._calculate,
        )

        self.assertEqual(
            (Calculator()(), self.recorder), (
                12,
                Recorder(saw=[EmittedDeprecation(object=unbound)]),
            ),
        )



import warnings

from twisted.trial.unittest import SynchronousTestCase

import regret


class Calculator(object):
    @regret.callable()
    def __call__(self):
        return 12

    @regret.callable()
    def calculate(self):
        return 12


@regret.callable()
def calculate():
    return 12


@regret.callable()
def add(x, y):
    return x + y


class TestRegret(SynchronousTestCase):
    def assertDeprecated(self, message, filename, fn, args=(), kwargs={}):
        # Sigh... assertWarns takes positional args positionally, instead of as
        # a sequence.
        return self.assertWarns(
            DeprecationWarning,
            message,
            filename,
            fn,
            *args,
            **kwargs
        )

    def test_function(self):
        result = self.assertDeprecated(
            message="'calculate' is deprecated.",
            filename=__file__,
            fn=calculate,
        )
        self.assertEqual(result, 12)

    def test_function_with_args(self):
        result = self.assertDeprecated(
            message="'add' is deprecated.",
            filename=__file__,
            fn=add,
            args=(9,),
            kwargs=dict(y=3),
        )
        self.assertEqual(result, 12)

    def test_method(self):
        calculator = Calculator()
        result = self.assertDeprecated(
            message="'Calculator.calculate' is deprecated.",
            filename=__file__,
            fn=calculator.calculate,
        )
        self.assertEqual(result, 12)

    def test_dunder_call(self):
        calculator = Calculator()
        result = self.assertDeprecated(
            message="'Calculator.__call__' is deprecated.",
            filename=__file__,
            fn=calculator,
        )
        self.assertEqual(result, 12)



from functools import wraps

from qualname import qualname
import attr

from regret import _warnings


@attr.s(eq=True, frozen=True)
class Deprecator(object):

    _emit = attr.ib(default=_warnings.emit)

    def emit_deprecation(self, **kwargs):
        self._emit(EmittedDeprecation(**kwargs))

    # -- Deprecatable objects --

    def callable(self):
        def deprecate(thing):
            @wraps(thing)
            def call_deprecated(*args, **kwargs):
                self.emit_deprecation(object=thing)
                return thing(*args, **kwargs)
            return call_deprecated
        return deprecate


@attr.s(eq=True, frozen=True, hash=True)
class EmittedDeprecation(object):

    _object = attr.ib()

    def message(self):
        return "{!r} is deprecated.".format(qualname(self._object))


_DEPRECATOR = Deprecator()
